import base64
import xml.etree.ElementTree as xml
from io import BytesIO
from pathlib import Path
from typing import Union

from PIL import Image


def _is_primitive(obj):
    """
    https://stackoverflow.com/questions/6391694/how-to-check-if-a-variables-type-is-primitive
    """
    return not hasattr(obj, "__dict__") and not isinstance(obj, list)


def base64img(img: Image.Image, img_suffix: str) -> str:
    """
    Convert image to base64
    """
    buffered = BytesIO()
    if img_suffix in [".JPG", ".jpg", ".JPEG", ".jpeg"]:
        format_ = "JPEG"
    elif img_suffix in [".PNG", ".png"]:
        format_ = "PNG"
    else:
        format_ = "PNG"
    img.save(buffered, format=format_)
    buffered.seek(0)
    img_byte = buffered.getvalue()
    encoded_string = base64.b64encode(img_byte).decode("utf-8")
    return encoded_string


def save_xml(output: Union[str, Path], xml_obj):
    """
    Save object to output file
    """
    tree = xml.ElementTree(xml_obj)
    xml.indent(tree, space="\t", level=0)
    with open(output, "w") as out:
        tree.write(out, encoding="unicode", method="xml")



import logging
from pathlib import Path
from typing import List, Optional, Union

from xmlobj.xmlmapping import XMLMixin

from pascal.draw_objects import DrawObjectsMixin
from pascal.exceptions import InconsistentAnnotation
from pascal.format_convertor import FormatConvertorMixin
from pascal.protocol_implementations import BBox, PascalObject, Size


class PascalAnnotation(DrawObjectsMixin, XMLMixin, FormatConvertorMixin):
    def __init__(
        self, filename: Union[Path, str], objects: List[PascalObject], size: Size
    ):
        self.filename = filename
        self.objects = objects
        self.size = size


def annotation_from_yolo(
    ann_path: Union[str, Path],
    img_w: int = 1,
    img_h: int = 1,
    label_map: Optional[dict] = None,
) -> PascalAnnotation:
    """
    Get annotation object from yolo ann file

    Parameters
    ----------
    ann_path: path to yolo annotation
    img_w: image width, px
    img_h: image height, px
    label_map: dict of labels and correspond ids, example:
        label_map = {0: "person", 1: "dog"}
    Returns
    -------
    PascalAnnotation object
    """
    objects = []
    has_label_map = label_map is not None
    with open(ann_path, "r") as f:
        lines = f.readlines()
    for line in lines:
        vals = line.split(" ")
        class_id = int(vals[0])
        x = float(vals[1])
        y = float(vals[2])
        dx = float(vals[3]) * 0.5
        dy = float(vals[4]) * 0.5
        x1 = x - dx
        x2 = x + dx
        y1 = y - dy
        y2 = y + dy
        x1 *= img_w
        x2 *= img_w
        y1 *= img_h
        y2 *= img_h
        if has_label_map:
            try:
                name = label_map[class_id]
            except KeyError:
                name = str(class_id)
                logging.warning(f"No id {class_id} in label map")
        else:
            name = str(class_id)
        box = BBox(x1, y1, x2, y2)
        pobj = PascalObject(name, box)
        objects.append(pobj)
    ann = PascalAnnotation(Path(ann_path).stem, objects, Size(img_w, img_h))
    if not isinstance(ann, PascalAnnotation):
        raise InconsistentAnnotation(f"Cannot make annotation from file: {ann_path}")
    return ann



from operator import itemgetter
from typing import List, Optional, Union

from typing_extensions import Protocol, runtime_checkable


@runtime_checkable
class BndBox(Protocol):
    xmin: Union[float, int]
    ymin: Union[float, int]
    xmax: Union[float, int]
    ymax: Union[float, int]


@runtime_checkable
class PascalObject(Protocol):
    name: str
    bndbox: BndBox


@runtime_checkable
class Size(Protocol):
    width: Union[float, int]
    height: Union[float, int]


@runtime_checkable
class PascalAnnotation(Protocol):
    objects: List[PascalObject] = []
    size: Optional[Size] = None
    filename: Optional[str] = None

    def __len__(self):
        return len(self.objects)

    def __getitem__(self, item):
        return self.objects[item]

    def filter_objects(self, names: List):
        """Filter objects by names"""
        ind = []
        for i, obj in enumerate(self.objects):
            if obj.name not in names:
                ind.append(i)
        if len(ind) == 0:
            self.objects = list()
        elif len(ind) < len(self.objects):
            self.objects = [itemgetter(*ind)(self.objects)]



from copy import deepcopy
from pathlib import Path
from typing import Optional, Union

from xmlobj import get_xml_obj
from xmlobj.xmlmapping import XMLMixin

from pascal.draw_objects import DrawObjectsMixin
from pascal.exceptions import InconsistentAnnotation, ParseException
from pascal.format_convertor import FormatConvertorMixin
from pascal.protocols import PascalAnnotation, Size
from pascal.utils import _is_primitive


def annotation_from_xml(
    file_path: Union[str, Path], attr_type_spec: Optional[dict] = None
) -> Union[PascalAnnotation, DrawObjectsMixin, FormatConvertorMixin, XMLMixin]:
    """
    Make annotation object from PascalVOC annotation file

    Parameters
    ----------
    file: path to xml file
    attr_type_spec: dict, optional
        specify attribute types to explicitly cast attribute values

    Returns
    -------
    Annotation object
    """
    try:
        obj = get_xml_obj(
            file_path,
            mixin_clsasses=[DrawObjectsMixin, FormatConvertorMixin],
            attr_type_spec=attr_type_spec,
        )
    except Exception as ex:
        raise ParseException(ex)
    if hasattr(obj, "object"):
        obj_ = getattr(obj, "object")
        if isinstance(obj_, list):
            objects = [deepcopy(obj) for obj in obj_]
        elif not _is_primitive(obj_):
            objects = [deepcopy(obj_)]
        else:
            raise ParseException("Cannot parse objects")
        setattr(obj, "objects", objects)
        delattr(obj, "object")
    if isinstance(obj, PascalAnnotation):
        if (obj.filename is None and len(obj.objects) == 0) or not isinstance(
            obj.size, Size
        ):
            raise InconsistentAnnotation(f"File {file_path} is not PascalVOCAnnotation")
    return obj



from pascal import annotation_from_xml


def test_yolo_convertation(yolo_data):
    """
    Проверить правильность конвертации
        Правильная yolo разметка получена через labelimg
    """
    for pairs in yolo_data:
        ann_file = pairs.get("xml_ann_file")
        ann = annotation_from_xml(ann_file)
        yolo_file = pairs.get("yolo_ann_file")
        with open(yolo_file, "r") as f:
            yolo_true_ann = f.readlines()
        yolo_true_ann = "".join(yolo_true_ann).strip()
        label_map = pairs.get("label_map")
        yolo_convert = ann.to_yolo(label_map, precision=6)
        assert yolo_convert == yolo_true_ann



import pytest
from xmlobj.xmlmapping import get_xml_obj

from pascal import annotation_from_xml
from pascal.exceptions import InconsistentAnnotation
from pascal.protocols import PascalObject


def check_attributes(obj, attr_params):
    attr_name = attr_params.get("attr_name")
    attr_val = attr_params.get("attr_val")
    assert hasattr(obj, attr_name)
    if isinstance(attr_val, list):
        for item in attr_val:
            check_attributes(getattr(obj, attr_name), item)
    else:
        attr_type = attr_params.get("dtype")
        assert isinstance(getattr(obj, attr_name), attr_type)


def test_valid(valid_annotations):
    """
    Проверить параметры валидных аннотаций
    """
    for ann_sample in valid_annotations:
        ann_file = ann_sample.get("file")
        ann = annotation_from_xml(ann_file)
        for attributes in ann_sample.get("attributes"):
            check_attributes(ann, attributes)


def test_invalid_annotations(invalid_ann_files):
    """
    Ошибка парсинга невалидных аннотаций
    """
    for file in invalid_ann_files:
        with pytest.raises(InconsistentAnnotation):
            annotation_from_xml(file)


def test_valid_objects(valid_objects):
    """
    Проверить параметры объектов
    """
    for obj_params in valid_objects:
        objects = get_xml_obj(obj_params.get("file"))
        for obj in objects.object:
            assert isinstance(obj, PascalObject)


def test_invalid_objects(invalid_objects):
    """
    Проверить невалидные объекты
    """
    for obj_params in invalid_objects:
        objects = get_xml_obj(obj_params.get("file"))
        for obj in objects.object:
            assert not isinstance(obj, PascalObject)



from dataclasses import dataclass
from typing import Union

from xmlobj.xmlmapping import XMLMixin

from pascal.protocols import BndBox as BndBoxProtocol
from pascal.protocols import PascalObject as PascalObjectProtocol
from pascal.protocols import Size as SizeProtocol


@dataclass
class BBox(BndBoxProtocol, XMLMixin):
    xmin: Union[float, int]
    ymin: Union[float, int]
    xmax: Union[float, int]
    ymax: Union[float, int]


@dataclass
class PascalObject(PascalObjectProtocol, XMLMixin):
    name: str
    bndbox: BBox


@dataclass
class Size(SizeProtocol, XMLMixin):
    width: Union[float, int]
    height: Union[float, int]



import hashlib
import logging
import random
from functools import lru_cache
from pathlib import Path
from typing import Optional

import PIL.ImageFont
from PIL import Image, ImageDraw
from transliterate import translit

from pascal.protocols import BndBox, PascalAnnotation
from pascal.protocols import PascalObject as PascalObjectProtocol

random.seed(32)

n_colors = 100
RAND_COLORS = [
    (random.randint(1, 255), random.randint(1, 255), random.randint(1, 255))
    for _ in range(n_colors)
]


@lru_cache
def get_name_hash(name: str) -> int:
    """
    https://stackoverflow.com/questions/16008670/how-to-hash-a-string-into-8-digits
    """
    return abs(int(hashlib.sha1(name.encode("utf-8")).hexdigest(), 16) % n_colors)


def is_relative_coords(box: BndBox) -> bool:
    return any(c < 1 for c in [box.xmin, box.ymin, box.xmax, box.ymax])


def _get_rect_coords(text_coord, d):
    x0 = text_coord[0] - d if text_coord[0] - 1 >= 0 else text_coord[0]
    y0 = text_coord[1] - d if text_coord[1] - 1 >= 0 else text_coord[1]
    x1 = text_coord[2] + d
    y1 = text_coord[3] + d
    return x0, y0, x1, y1


class DrawObjectsMixin(PascalAnnotation):
    """
    Draw objects
    """

    _font_path = str(Path(__file__).parent / "fonts/arialmt.ttf")

    def draw_boxes(
        self,
        image: Image.Image,
        width: int = 5,
        color: Optional[tuple[int, int, int]] = None,
        fontsize: int = 10,
        font_path: str = None,
        language_code: Optional[str] = None,
    ) -> Image.Image:
        """
        Draw bounding boxes and obj names

        Parameters
        ----------
        image: Pillow image
        width: The line width, in pixels.
            specify attribute types to explicitly cast attribute values
        color: bounding box color
        fontsize: requested font size, in pixels
        font_path: path to font
        language_code: str language code if transliteration needs

        Returns
        -------
        Copy of image with rendered boxes
        """
        img_copy = image.copy()
        img_width = img_copy.width
        img_height = img_copy.height
        img_draw = ImageDraw.Draw(img_copy)
        set_color = color is None
        if font_path is not None:
            font = PIL.ImageFont.truetype(font_path, size=fontsize)
        else:
            font = PIL.ImageFont.truetype(self._font_path, size=fontsize)
        for obj in self.objects:
            if not isinstance(obj, PascalObjectProtocol):
                logging.warning("Annotation has object which is not PascalObject")
                continue
            if set_color:
                hash_id = get_name_hash(str(obj.name))
                color = RAND_COLORS[hash_id]
            # draw rectangle
            if is_relative_coords(obj.bndbox):
                p1 = (
                    float(obj.bndbox.xmin * img_width),
                    float(obj.bndbox.ymin * img_height),
                )
                p2 = (
                    float(obj.bndbox.xmax * img_width),
                    float(obj.bndbox.ymax * img_height),
                )
            else:
                p1 = (float(obj.bndbox.xmin), float(obj.bndbox.ymin))
                p2 = (float(obj.bndbox.xmax), float(obj.bndbox.ymax))
            img_draw.rectangle((p1, p2), outline=color, width=width)
            # draw text
            text_coord = (
                float(p1[0] + width + 1),
                float(p1[1] + width + 1),
            )
            if language_code is not None:
                obj_name = str(translit(obj.name, language_code, reversed=True))
            else:
                obj_name = str(obj.name)
            text_box = img_draw.textbbox(text_coord, obj_name, font=font)
            rect_coords = _get_rect_coords(text_box, width // 2)
            img_draw.rectangle(rect_coords, fill=(32, 32, 28))
            img_draw.text(text_coord, obj_name, align="left", font=font)
        return img_copy



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="pascal_voc",
    version="2.1.1",
    author="Alexander Barmin",
    author_email="barmin1@mail.ru",
    description="Tool to work with annotation formats",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(
        include=["pascal"],
        exclude=["*.tests", "*.tests.*", "tests.*", "tests"],
    ),
    data_files=[("pascal", ["pascal/fonts/arialmt.ttf"])],
    install_requires=[
        "Pillow>=8.1.0",
        "transliterate>=1.10.2",
        "xmlobj==1.2.0",
        "typing_extensions==4.7.1",
    ],
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import logging
from pathlib import Path
from typing import List, Union

from PIL import Image

from pascal.exceptions import InconsistentAnnotation
from pascal.protocols import PascalAnnotation, PascalObject, Size
from pascal.utils import _is_primitive, base64img


def get_shapes(obj_data) -> List[dict]:
    if len(obj_data) == 0:
        return []
    shapes = []
    for obj in obj_data:
        if not isinstance(obj, PascalObject):
            logging.warning("Annotation has object which is not PascalObject")
            continue
        label = obj.name
        points = [
            [obj.bndbox.xmin, obj.bndbox.ymin],
            [obj.bndbox.xmax, obj.bndbox.ymin],
            [obj.bndbox.xmax, obj.bndbox.ymax],
            [obj.bndbox.xmin, obj.bndbox.ymax],
        ]
        obj_flags = {}
        for k, v in obj.__dict__.items():
            if _is_primitive(v) and k != "name":
                obj_flags[k] = v
            if isinstance(v, list):
                shapes.extend(get_shapes(v))
        shape = dict(
            label=label,
            points=points,
            group_id=None,
            shape_type="polygon",
            flags=obj_flags,
        )
        shapes.append(shape)
    return shapes


class FormatConvertorMixin(PascalAnnotation):
    def to_yolo(self, labels_map: dict, precision: int = 3) -> str:
        """
        Convert annotation to yolo format str
        Annotation object must have size attribute, which contains fields: width, height

        Parameters
        ----------
        labels_map: dict of label ids
            {"person": 0, "cat": 1, "dog": 2}
        precision: int, coord precision

        Returns
        -------
        Yolo format annotation str
        """
        if not isinstance(self.size, Size):
            raise InconsistentAnnotation(
                "Incorrect size. Size must have width and height attributes"
            )

        objects = []
        for obj in self:
            if not isinstance(obj, PascalObject):
                logging.warning("Annotation has object which is not PascalObject")
                continue
            try:
                label = labels_map[obj.name]
            except KeyError:
                logging.warning(f"No label {obj.name} in label map. Skip object")
                continue
            dx = float(obj.bndbox.xmax - obj.bndbox.xmin)
            dy = float(obj.bndbox.ymax - obj.bndbox.ymin)
            x = obj.bndbox.xmin + dx * 0.5
            y = obj.bndbox.ymin + dy * 0.5
            dx /= self.size.width
            dy /= self.size.height
            x /= self.size.width
            y /= self.size.height
            s = f"{label} {x:.{precision}f} {y:.{precision}f} {dx:.{precision}f} {dy:.{precision}f}"
            objects.append(s)
        return "\n".join(objects)

    def to_labelme(
        self,
        img_path: Union[str, Path] = None,
        save_img_data: bool = False,
        label_me_version: str = "5.3.0",
    ) -> dict:
        """
        Convert annotation to labelme format

        Parameters
        ----------
        img_path: path to image
        save_img_data: if true store encoded image in output dict
        label_me_version: version of labelme app

        Returns
        -------
        labelme annotation dict which can be saved as json
        """
        if not img_path.exists():
            raise FileNotFoundError(f"No such file: {img_path}")

        encoded_string = None
        if save_img_data:
            img = Image.open(img_path)
            encoded_string = base64img(img, img_path.suffix)

        shapes = get_shapes(self.objects)

        res = dict(
            version=label_me_version,
            flags={},
            shapes=shapes,
            imagePath=str(img_path),
            imageData=encoded_string,
            imageHeight=self.size.height,
            imageWidth=self.size.width,
        )
        return res
